# This file is a part of the AnyBlok project
#
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in
from anyblok.column import Integer as Int, String as Str
from anyblok.migration import MigrationException, DropSchema
from contextlib import contextmanager
from sqlalchemy import (
    MetaData, Table, Column, Integer, String, CheckConstraint, ForeignKey,
    text)
from anyblok import Declarations
from sqlalchemy.exc import IntegrityError
from anyblok.config import Configuration
from .conftest import init_registry, drop_database, create_database
from anyblok.common import naming_convention


@pytest.fixture(scope="module")
def clean_db(request, configuration_loaded):
    def clean():
        url = Configuration.get('get_url')()
        drop_database(url)
        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)

    request.addfinalizer(clean)


@contextmanager
def cnx(registry):
    cnx = registry.migration.conn
    try:
        yield cnx
    except Exception:
        cnx.execute(text("rollback"))
        raise


def add_in_registry():
    register = Declarations.register
    Model = Declarations.Model

    @register(Model)
    class Test:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Str()

    @register(Model)
    class TestUnique:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Str(unique=True)

    @register(Model)
    class TestIndex:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Str(index=True)

    if not sgdb_in(['MySQL', 'MariaDB']):
        @register(Model)
        class TestCheck:
            __db_schema__ = 'test_db_schema'

            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheck, cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name='test'),)

    @register(Model)
    class TestFKTarget:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)

    @register(Model)
    class TestFK:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer'))

    @register(Model)
    class TestFK2:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer').options(
                ondelete='cascade'))


@pytest.fixture()
def registry(request, clean_db, bloks_loaded):
    registry = init_registry(add_in_registry)

    def rollback():
        for table in ('test', 'testfk', 'testunique', 'testfk2',
                      'testfktarget', 'testindex', 'testcheck'):
            try:
                registry.migration.table(table, schema='test_db_schema').drop()
            except MigrationException:
                pass

        registry.close()

    request.addfinalizer(rollback)
    return registry


class TestMigrationDbSchema:

    def test_add_schema(self, registry):
        registry.migration.schema().add('other_schema')
        registry.migration.schema('other_schema')
        registry.migration.schema('other_schema').drop()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="Can't create empty table")
    def test_add_table_from_schema(self, registry):
        schema = registry.migration.schema('test_db_schema')
        schema.table().add('test2')
        table = schema.table('test2')
        assert table.schema == 'test_db_schema'

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="Can't create empty table")
    def test_add_table(self, registry):
        registry.migration.table(schema='test_db_schema').add('test2')
        table = registry.migration.table('test2', schema='test_db_schema')
        assert table.schema == 'test_db_schema'

    def test_add_column(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_unique_constraint(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.unique('unique_constraint').add(t.column('other'))
        registry.Test.insert(other='One entry')
        with pytest.raises(IntegrityError):
            registry.Test.insert(other='One entry')

    def test_drop_schema(self, registry):
        schema = registry.migration.schema('test_db_schema')
        schema.table('test').drop()
        schema.table('testfk').drop()
        schema.table('testunique').drop()
        schema.table('testfk2').drop()
        schema.table('testfktarget').drop()
        schema.table('testindex').drop()
        if not sgdb_in(['MySQL', 'MariaDB']):
            schema.table('testcheck').drop()

        schema.drop()
        with pytest.raises(MigrationException):
            registry.migration.schema('test_db_schema')

        # MySQL waiting that this schema exist for the next test
        registry.migration.schema().add('test_db_schema')

    def test_drop_table(self, registry):
        registry.migration.table('test', schema='test_db_schema').drop()
        with pytest.raises(MigrationException):
            registry.migration.table('Test', schema='test_db_schema')

    def test_drop_column(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.column('other').drop()
        with pytest.raises(MigrationException):
            t.column('other')

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="Can't rename schema #122")
    def test_alter_schema_name(self, registry):
        registry.migration.schema('test_db_schema').alter(name='test_db_other')
        with pytest.raises(MigrationException):
            registry.migration.schema('test_db_schema')

        registry.migration.schema('test_db_other')
        registry.migration.schema('test_db_other').table('test')

    def test_alter_table_name(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.alter(name='othername')
        registry.migration.table('othername', schema='test_db_schema')
        with pytest.raises(MigrationException):
            registry.migration.table('test')

    def test_alter_column_name(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.column('other').alter(name='othername')
        t.column('othername')
        with pytest.raises(MigrationException):
            t.column('other')

    def test_index(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.index().add(t.column('integer'))
        t.index(t.column('integer')).drop()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="FIXME: can't create foreign key")
    def test_alter_column_foreign_key(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.foreign_key('my_fk').add(
            ['other'], [registry.System.Blok.name])
        t.foreign_key('my_fk').drop()

    def test_constraint_unique(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        t.unique(name='test_unique_constraint').add(t.column('other'))
        t.unique(name='test_unique_constraint').drop()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="Can't drop check constraint issue #93")
    def test_constraint_check(self, registry):
        t = registry.migration.table('test', schema='test_db_schema')
        Test = registry.Test
        t.check('test').add(Test.other != 'test')
        # particuliar case of check constraint
        t.check('anyblok_ck_test__test').drop()

    def test_detect_schema_added(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.TestFK.__table__.drop(bind=conn)
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestFK2.__table__.drop(bind=conn)
            registry.TestFKTarget.__table__.drop(bind=conn)
            registry.TestIndex.__table__.drop(bind=conn)
            if not sgdb_in(['MySQL', 'MariaDB']):
                registry.TestCheck.__table__.drop(bind=conn)

            conn.execute(DropSchema('test_db_schema'))

        report = registry.migration.detect_changed(schema_only=True)
        assert report.log_has("Add schema test_db_schema")
        report.apply_change()
        report = registry.migration.detect_changed(schema_only=True)
        assert not report.log_has("Add schema test_db_schema")

    def test_detect_table_added(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add table test_db_schema.test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has("Add table test_db_schema.test")

    def test_detect_column_added(self, registry):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has("Add test.other")

    def test_detect_table_removed(self, registry):
        with cnx(registry) as conn:
            Table(
                'test2', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
                schema='test_db_schema'
            ).create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test_db_schema.test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test_db_schema.test2")

    def test_detect_column_removed(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")

    def test_detect_nullable(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), nullable=False),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    def test_detect_add_index_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestIndex.__table__.drop(bind=conn)
            registry.TestIndex.__table__ = Table(
                'testindex', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                schema='test_db_schema'
            )
            registry.TestIndex.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add index constraint on testindex (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add index constraint on testindex (other)"))

    def test_detect_add_column_with_index_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestIndex.__table__.drop(bind=conn)
            registry.TestIndex.__table__ = Table(
                'testindex', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.TestIndex.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add testindex.other")
        assert report.log_has("Add index constraint on testindex (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Add testindex.other"))
        assert not(report.log_has(
            "Add index constraint on testindex (other)"))

    def test_detect_drop_anyblok_index(self, registry):
        with cnx(registry) as conn:
            conn.execute(text(
                """CREATE INDEX anyblok_ix_test__other
                   ON test_db_schema.test (other);"""))
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop index anyblok_ix_test__other on test_db_schema.test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has(
                "Drop index anyblok_ix_test__other on test_db_schema.test"))

    def test_detect_type(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    def test_detect_primary_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, nullable=False),
                Column('other', String(64), primary_key=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    def test_detect_add_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.TestFK.__table__.drop(bind=conn)
            registry.TestFK.__table__ = Table(
                'testfk', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
                schema='test_db_schema'
            )
            registry.TestFK.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Add Foreign keys on (testfk.other) => "
            "(test_db_schema.testfktarget.integer)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add Foreign keys on (testfk.other) => "
            "(test_db_schema.testfktarget.integer)")

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not change fk")
    def test_detect_foreign_key_options_changed(self, registry):
        with cnx(registry) as conn:
            registry.TestFK2.__table__.drop(bind=conn)
            meta = MetaData()
            meta._add_table(
                'testfktarget', None, registry.TestFKTarget.__table__)
            registry.TestFK2.__table__ = Table(
                'testfk2', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', Integer, ForeignKey('testfktarget.integer')),
                schema='test_db_schema'
            )
            registry.TestFK2.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on testfk2.other => "
            "test_db_schema.testfktarget.integer")
        assert report.log_has(
            "Add Foreign keys on (testfk2.other) => "
            "(test_db_schema.testfktarget.integer)")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="FIXME: can't create foreign key")
    def test_detect_drop_anyblok_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)
            # anyblok_fk_test__other_on_system_blok__name

        report = registry.migration.detect_changed()

        message = (
            "Drop Foreign keys on test.other => %ssystem_blok.name"
        ) % ('dbo.' if sgdb_in(['MsSQL']) else '')
        assert report.log_has(message)
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has(message)

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="FIXME: can't create foreign key")
    def test_detect_drop_column_with_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64), ForeignKey('system_blok.name')),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        message = (
            "Drop Foreign keys on test.other2 => %ssystem_blok.name"
        ) % ('dbo.' if sgdb_in(['MsSQL']) else '')
        assert report.log_has(message)
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has(message)

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_unique_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                schema='test_db_schema'
            )
            registry.TestUnique.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add unique constraint on testunique (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add unique constraint on testunique (other)"))

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_column_with_unique_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.TestUnique.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add unique constraint on testunique (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add unique constraint on testunique (other)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop constraint anyblok_uq_test__other on test_db_schema.test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Drop constraint anyblok_uq_test__other on test_db_schema.test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_add_check_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestCheck.__table__.drop(bind=conn)
            registry.TestCheck.__table__ = Table(
                'testcheck', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.TestCheck.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Add check constraint anyblok_ck_testcheck__test "
            "on test_db_schema.testcheck")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add check constraint anyblok_ck_testcheck__test "
            "on test_db_schema.testcheck")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64),
                       CheckConstraint("other != 'test'", name='check')),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop check constraint anyblok_ck_test__check "
            "on test_db_schema.test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop check constraint anyblok_ck_test__check "
            "on test_db_schema.test"))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok


class TestBlok(Blok):

    version = "1.0.0"

    called_methods = []

    def pre_migration(self, latest_version):
        self.__class__.called_methods.append("pre_migration")

    def update(self, latest_version):
        self.__class__.called_methods.append("update")

    def post_migration(self, latest_version):
        self.__class__.called_methods.append("post_migration")

    def update_demo(self, latest_version):
        self.__class__.called_methods.append("update_demo")

    def uninstall(self):
        self.__class__.called_methods.append("uninstall")

    def uninstall_demo(self):
        self.__class__.called_methods.append("uninstall_demo")



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.field import FieldException
from anyblok.relationship import RelationShip, RelationShipList
from anyblok import Declarations
from anyblok.column import Integer
from anyblok.relationship import One2Many, Many2One
from .conftest import init_registry_with_bloks


register = Declarations.register
Model = Declarations.Model


class OneRelationShip(RelationShip):
    pass


class OneModel:
    __tablename__ = 'test'
    __registry_name__ = 'One.Model'


class MockRegistry:

    InstrumentedList = []


class TestRelationShip:

    def test_forbid_instance(self):
        with pytest.raises(FieldException):
            RelationShip(model=OneModel)

    def test_must_have_a_model(self):
        OneRelationShip(model=OneModel)
        with pytest.raises(FieldException):
            OneRelationShip()


class List(RelationShipList, list):
    def relationship_field_append_value(*a, **kw):
        return True

    def relationship_field_remove_value(*a, **kw):
        return True


class TestRelationShipList:

    def test_append(self):
        list_ = List()
        list_.append(1)
        assert list_ == [1]

    def test_extend(self):
        list_ = List()
        list_.extend([1, 2])
        assert list_ == [1, 2]

    def test_insert(self):
        list_ = List()
        list_.insert(0, 1)
        assert list_ == [1]
        list_.insert(0, 2)
        assert list_ == [2, 1]

    def test_pop(self):
        list_ = List()
        list_.extend([1, 2])
        assert list_ == [1, 2]
        list_.pop(0)
        assert list_ == [2]
        list_.pop(0)
        assert list_ == []

    def test_remove(self):
        list_ = List()
        list_.extend([3, 2, 1])
        assert list_ == [3, 2, 1]
        list_.remove(1)
        assert list_ == [3, 2]

    def test_clear(self):
        list_ = List()
        list_.extend([3, 2, 1])
        assert list_ == [3, 2, 1]
        list_.clear()
        assert list_ == []


def multi_parallel_foreign_key_with_definition():

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        id = Integer(primary_key=True)
        address_1_id = Integer(foreign_key=Model.Address.use('id'))
        address_2_id = Integer(foreign_key=Model.Address.use('id'))
        address_1 = Many2One(model=Model.Address,
                             column_names=['address_1_id'])
        address_2 = Many2One(model=Model.Address,
                             column_names=['address_2_id'])

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model=Model.Person)


def multi_parallel_foreign_key_auto_detect():

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        persons = One2Many(model='Model.Person')

    @register(Model)
    class Person:

        id = Integer(primary_key=True)
        address_1 = Many2One(model=Model.Address)
        address_2 = Many2One(model=Model.Address)


@pytest.fixture(
    scope="class",
    params=[
        multi_parallel_foreign_key_with_definition,
        multi_parallel_foreign_key_auto_detect,
    ]
)
def registry_relationship_multiple_foreign_keys(request, bloks_loaded):
    registry = init_registry_with_bloks(
        [], request.param)
    request.addfinalizer(registry.close)
    return registry


class TestComplexeRelationShipCase:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_relationship_multiple_foreign_keys):
        transaction = registry_relationship_multiple_foreign_keys.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_with_multi_parallel_foreign_key_with_definition(
        self, registry_relationship_multiple_foreign_keys
    ):

        registry = registry_relationship_multiple_foreign_keys
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            address_1=address_1, address_2=address_2)
        assert address_1.persons == [person]
        assert address_2.persons == [person]



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in
from anyblok import Declarations
from anyblok.config import Configuration
from sqlalchemy.exc import IntegrityError
from anyblok.field import FieldException
from anyblok.column import (
    Integer, String, BigInteger, Float, Decimal, Boolean, DateTime, Date, Time,
    Sequence, Email, UUID, URL, Country, Color, PhoneNumber, Selection,
    TimeStamp)
from anyblok.relationship import Many2One, ordering_list
from sqlalchemy import ForeignKeyConstraint, inspect
from .conftest import init_registry_with_bloks, init_registry, reset_db


COLUMNS = [
    Integer,
    BigInteger,
    String,
    Boolean,
    Date,
    Time,
    Email,
    UUID,
    Selection,
]


if not sgdb_in(['MySQL', 'MariaDB', 'MsSQL']):
    COLUMNS.extend([DateTime, TimeStamp])


try:
    import colour  # noqa
    COLUMNS.append(Color)
except Exception:
    pass

try:
    import furl  # noqa
    if not sgdb_in(['MySQL', 'MariaDB', 'MsSQL']):
        COLUMNS.append(URL)
except Exception:
    pass


try:
    import phonenumbers  # noqa
    COLUMNS.append(PhoneNumber)
except Exception:
    pass

try:
    import pycountry  # noqa
    COLUMNS.append(Country)
except Exception:
    pass


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


@pytest.fixture(
    scope="class",
    params=[
        ('prefix', 'suffix'),
        ('', ''),
    ]
)
def db_schema(request, bloks_loaded):
    Configuration.set('prefix_db_schema', request.param[0])
    Configuration.set('suffix_db_schema', request.param[1])

    def rollback():
        Configuration.set('prefix_db_schema', '')
        Configuration.set('suffix_db_schema', '')

    request.addfinalizer(rollback)


def _complete_many2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many="persons", nullable=False)


def _complete_many2one_with_ordered_one2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many=(
                               "persons",
                               dict(
                                   order_by="ModelPerson.name",
                                   collection_class=ordering_list('name'),
                               ),
                           ),
                           nullable=False)


def _complete_many2one_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2o_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2o_schema'

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many="persons", nullable=False)


def _complete_many2one_with_different_schema1(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2o_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_other_m2o_schema'

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many="persons", nullable=False)


def _complete_many2one_with_different_schema2(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2o_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many="persons", nullable=False)


def _minimum_many2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


def _many2one_to_polymorphic_models_1(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()
        type_entity = String(default='shipping')

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Address, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model.Address, tablename=Model.Address)
    class Shipping(Model.Address):

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Shipping, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'shipping',
            })
            return mapper_args

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address, one2many="persons")


def _many2one_to_polymorphic_models_2(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()
        type_entity = String(default='shipping')

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Address, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model.Address)
    class Shipping(Model.Address):
        id = Integer(primary_key=True, foreign_key=Model.Address.use('id'))

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Shipping, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'shipping',
            })
            return mapper_args

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address, one2many="persons")


def _minimum_many2one_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2o_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2o_schema'

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


def _many2one_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model='Model.Address')


def many2one_on_mapping_model_and_column_1(**kwargs):

    @register(Model, tablename="x")
    class Address:

        id = Integer(primary_key=True, db_column_name="x1")
        street = String(db_column_name="x2")
        zip = String(db_column_name="x3")
        city = String(db_column_name="x4")

    @register(Model, tablename="y")
    class Person:

        name = String(primary_key=True, db_column_name="y1")
        address = Many2One(model=Model.Address)


def many2one_on_mapping_model_and_column_2(**kwargs):

    @register(Model, tablename="x")
    class Address:

        id = Integer(primary_key=True, db_column_name="x1")
        street = String(db_column_name="x2")
        zip = String(db_column_name="x3")
        city = String(db_column_name="x4")

    @register(Model, tablename="y")
    class Person:

        name = String(primary_key=True, db_column_name="y1")
        address_id = Integer(db_column_name="y2",
                             foreign_key=Model.Address.use('id'))
        address = Many2One(model=Model.Address)


@pytest.fixture(
    scope="class",
    params=[
        (_complete_many2one, 'id_of_address', True),
        (_complete_many2one_with_ordered_one2many, 'id_of_address', True),
        (_complete_many2one_with_schema, 'id_of_address', True),
        (_complete_many2one_with_different_schema1, 'id_of_address', True),
        (_complete_many2one_with_different_schema2, 'id_of_address', True),
        (_minimum_many2one, 'address_id', False),
        (_many2one_to_polymorphic_models_1, 'address_id', True),
        (_many2one_to_polymorphic_models_2, 'address_id', True),
        (_minimum_many2one_with_schema, 'address_id', False),
        (_many2one_with_str_model, 'address_id', False),
        (many2one_on_mapping_model_and_column_1, 'address_id', False),
        (many2one_on_mapping_model_and_column_2, 'address_id', False),
    ]
)
def registry_many2one(request, bloks_loaded, db_schema):
    reset_db()
    registry = init_registry_with_bloks(
        [], request.param[0])
    request.addfinalizer(registry.close)
    return registry, request.param[1], request.param[2]


@pytest.mark.relationship
class TestMany2One:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_many2one):
        transaction = registry_many2one[0].begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_ok(self, request, registry_many2one):
        registry, column_name, has_one2many = registry_many2one
        address_exist = hasattr(registry.Person, 'address')
        assert address_exist
        column_name_exist = hasattr(registry.Person, column_name)
        assert column_name_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        if has_one2many:
            assert address.persons == [person]
            person.delete()  # issue 68
            address.expire('persons')

    def test_autodoc(self, registry_many2one):
        registry, _, _ = registry_many2one
        registry.loaded_namespaces_first_step['Model.Person'][
            'address'].autodoc_get_properties()

    def test_expire_field(self, registry_many2one):
        registry, column_name, _ = registry_many2one
        id_of_address = registry.expire_attributes['Model.Person'][
            column_name]
        assert ('address',) in id_of_address


def _auto_detect_type(ColumnType=None, **kwargs):

    @register(Model)
    class Address:

        id = ColumnType(primary_key=True, **kwargs)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


@pytest.fixture(scope="class", params=COLUMNS)
def registry_many2one_auto_detect(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks(
        [], _auto_detect_type, ColumnType=request.param)
    request.addfinalizer(registry.close)
    return registry


@pytest.mark.relationship
class TestMany2OneAutoDetectColumn:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_many2one_auto_detect):
        transaction = registry_many2one_auto_detect.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_check_autodetect_type(self, registry_many2one_auto_detect):
        registry = registry_many2one_auto_detect
        assert(
            str(registry.Address.id.property.columns[0].type) ==
            str(registry.Person.address_id.property.columns[0].type)
        )


def _many2one_with_same_name_for_column_names(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           column_names="address")


def _minimum_many2one_without_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One()


def _two_remote_primary_keys(**kwargs):

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = Many2One(model=Model.Test)


@pytest.mark.relationship
class TestMany2OneOld:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_2_many2one(self):
        with pytest.raises(FieldException):
            self.init_registry(_many2one_with_same_name_for_column_names)

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason='ISSUE #89')
    def test_minimum_many2one_on_sequence(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                seq = Sequence(primary_key=True, formater="V-{seq}")

            @register(Model)
            class Test2:

                seq = Sequence(primary_key=True)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        test2 = registry.Test2.insert(test=test)
        assert test is test2.test

    def test_minimum_many2one_without_model(self):
        with pytest.raises(FieldException):
            self.init_registry(_minimum_many2one_without_model)

    def test_same_model(self):
        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        assert t2.parent is t1
        assert t1.children[0] is t2

    def test_same_model_pk_by_inherit(self):
        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class Test:  # noqa

                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        assert t2.parent is t1
        assert t1.children[0] is t2

    def test_with_primary_key_eq_False(self):
        def add_in_registry():

            @register(Model)
            class Test1:

                id = Integer(primary_key=True)
                id2 = Integer(primary_key=False)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                parent = Many2One(model='Model.Test1')

        self.init_registry(add_in_registry)

    def test_same_model_pk_by_mixin(self):
        def add_in_registry():

            @register(Mixin)
            class MTest:

                id = Integer(primary_key=True)

            @register(Model)  # noqa
            class Test(Mixin.MTest):

                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        assert t2.parent is t1
        assert t1.children[0] is t2

    def test_add_unique_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, unique=True)

        registry = self.init_registry(add_in_registry)
        address = registry.Address.insert()
        registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        with pytest.raises(IntegrityError):
            registry.Person.insert(name="Other", address=address)

    def test_add_index_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, index=True)

        registry = self.init_registry(add_in_registry)
        inspector = inspect(registry.session.connection())
        indexes = inspector.get_indexes(registry.Person.__tablename__)
        assert len(indexes) == 1

    def test_add_primary_keys_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, primary_key=True)

        registry = self.init_registry(add_in_registry)
        inspector = inspect(registry.session.connection())
        constraint = inspector.get_pk_constraint(registry.Person.__tablename__)
        pks = constraint['constrained_columns']
        assert 'address_id' in pks

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                @classmethod
                def define_table_args(cls):
                    table_args = super(Test2, cls).define_table_args()
                    return table_args + (ForeignKeyConstraint(
                        [cls.test_id, cls.test_id2], ['test.id', 'test.id2']),)

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test,
                                remote_columns=('id', 'id2'),
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 is test2.test_id2

    def test_complet_with_multi_foreign_key_without_constraint(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test,
                                remote_columns=('id', 'id2'),
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 is test2.test_id2

    def test_with_multi_foreign_key_on_existing_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 is test2.test_id2

    def test_with_multi_foreign_key_on_existing_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                other_test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                other_test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.other_test_id
        assert test.id2 is test2.other_test_id2

    def test_with_multi_foreign_key_on_unexisting_column(self):
        registry = self.init_registry(_two_remote_primary_keys)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 is test2.test_id2

    def test_many2one_with_multi_fk_expire_field(self):
        registry = self.init_registry(_two_remote_primary_keys)
        assert('test',) in registry.expire_attributes['Model.Test2']['test_id']
        assert ('test',) in registry.expire_attributes[
            'Model.Test2']['test_id2']

    def test_with_multi_foreign_key_on_unexisting_named_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test,
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 is test2.test_id2

    def test_with_multi_foreign_key_on_unexisting_named_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = Integer(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, column_names=(
                    'other_test_id', 'other_test_id2'))

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_m2o_in_mixin(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Mixin)
            class MTest:

                test = Many2One(model=Model.Test)

            @register(Model)
            class Test1(Mixin.MTest):

                id = Integer(primary_key=True)

            @register(Model)
            class Test2(Mixin.MTest):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        test1 = registry.Test1.insert(test=test)
        test2 = registry.Test2.insert(test=test)
        assert test1.test_id == test2.test_id

    def test_delete_m2o_without_fk_options_on_delete_cascade(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class TestM2O:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, nullable=True)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        testM2O = registry.TestM2O.insert(test=test)
        assert testM2O.test is test
        with pytest.raises(IntegrityError):
            test.delete()

    def test_delete_m2o_with_fk_options_on_delete_cascade(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class TestM2O:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, nullable=True,
                                foreign_key_options={'ondelete': 'cascade'})

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        testM2O = registry.TestM2O.insert(test=test)
        assert testM2O.test is test
        test.delete()
        assert not (registry.TestM2O.query().count())

    def test_2_simple_many2one_on_the_same_model(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1 = Many2One(model=Model.Address)
                address_2 = Many2One(model=Model.Address)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1=address_1, address_2=address_2)
        assert person.address_1 is address_1
        assert person.address_2 is address_2
        assert person.address_1 is not person.address_2

    def test_many2one_with_order_list_on_one2many(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                id = Integer(primary_key=True)
                name = String()
                address = Many2One(
                    model=Model.Address,
                    one2many=(
                        'persons',
                        dict(
                            order_by="ModelPerson.name",
                            collection_class=ordering_list('name'),
                        )
                    )
                )

        registry = self.init_registry(add_in_registry)
        address = registry.Address.insert()
        p3 = registry.Person.insert(name='test 3', address=address)
        p1 = registry.Person.insert(name='test 1', address=address)
        p4 = registry.Person.insert(name='test 4', address=address)
        p2 = registry.Person.insert(name='test 2', address=address)
        assert address.persons == [p1, p2, p3, p4]
        assert address.persons.ordering_attr == 'name'



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2021 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.testing import sgdb_in
from anyblok.model.factory import ViewFactory
from anyblok.model.exceptions import ViewException
from anyblok import Declarations
from sqlalchemy.sql import select, expression, union
from anyblok.column import Integer, String
from anyblok.relationship import Many2One
import pytest
from .conftest import init_registry_with_bloks, reset_db

register = Declarations.register
Model = Declarations.Model


def simple_view():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            T2 = cls.anyblok.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)


@pytest.fixture(scope="class")
def registry_simple_view(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks([], simple_view)
    request.addfinalizer(registry.close)
    registry.T1.insert(code='test1', val=1)
    registry.T2.insert(code='test1', val=2)
    registry.T1.insert(code='test2', val=3)
    registry.T2.insert(code='test2', val=4)
    return registry


class TestSimpleView:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_simple_view):
        transaction = registry_simple_view.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_has_a_mapper(self, registry_simple_view):
        registry = registry_simple_view
        assert registry.TestView.__mapper__ is not None

    def test_ok(self, registry_simple_view):
        registry = registry_simple_view
        TestView = registry.TestView
        v1 = TestView.query().filter(TestView.code == 'test1').first()
        v2 = TestView.query().filter(TestView.code == 'test2').first()
        assert v1.val1 == 1
        assert v1.val2 == 2
        assert v2.val1 == 3
        assert v2.val2 == 4

    def test_view_update_method(self, registry_simple_view):
        registry = registry_simple_view
        with pytest.raises(AttributeError):
            registry.TestView.execute_sql_statement(
                registry.TestView.update_sql_statement().values({'val2': 3}))

    def test_view_delete_method(self, registry_simple_view):
        registry = registry_simple_view
        with pytest.raises(AttributeError):
            registry.TestView.execute_sql_statement(
                registry.TestView.delete_sql_statement())


def view_with_relationship():

    @register(Model)
    class Rs:
        id = Integer(primary_key=True)

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()
        rs_id = Integer(foreign_key=Model.Rs.use('id'))
        rs = Many2One(model=Model.Rs, column_names='rs_id')

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()
        rs = Many2One(model=Model.Rs)

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            T2 = cls.anyblok.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T1.rs_id.label('rs_id'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)


@pytest.fixture(scope="class")
def registry_view_with_relationship(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks(
        [], view_with_relationship)
    rs1 = registry.Rs.insert(id=1)
    rs2 = registry.Rs.insert(id=2)
    registry.T1.insert(code='test1', val=1, rs=rs1)
    registry.T2.insert(code='test1', val=2)
    registry.T1.insert(code='test2', val=3, rs=rs2)
    registry.T2.insert(code='test2', val=4)
    yield registry
    registry.close()


class TestViewWithRelationShip:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_view_with_relationship):
        transaction = registry_view_with_relationship.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_ok(self, registry_view_with_relationship):
        registry = registry_view_with_relationship
        TestView = registry.TestView
        v1 = TestView.query().filter(TestView.code == 'test1').first()
        v2 = TestView.query().filter(TestView.code == 'test2').first()
        assert v1.val1 == 1
        assert v1.val2 == 2
        assert v1.rs.id == 1
        assert v2.val1 == 3
        assert v2.val2 == 4
        assert v2.rs.id == 2


def view_with_relationship_on_self():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()
        parent = Many2One(model='Model.T1')

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()
        parent = Many2One(model='Model.TestView')

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            TP = cls.anyblok.T1.aliased()
            T2 = cls.anyblok.T2
            subquery = union(
                select([
                    T1.code.label('code'),
                    TP.code.label('parent_code')]
                ).where(T1.parent_id == TP.id),
                select([
                    T1.code.label('code'),
                    expression.literal_column("null as parent_code")
                ]).where(T1.parent_id.is_(None))
            ).alias()
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2'),
                            subquery.c.parent_code.label('parent_code')])
            query = query.where(subquery.c.code == T1.code)
            query = query.where(subquery.c.code == T2.code)
            return query


def view_with_relationship_on_self_2():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()
        parent = Many2One(model='Model.T1')

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer(primary_key=True)
        val2 = Integer()
        parent = Many2One(model='Model.TestView')

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            TP = cls.anyblok.T1.aliased()
            T2 = cls.anyblok.T2
            subquery = union(
                select([
                    T1.code.label('code'),
                    TP.id.label('parent_val1'),
                    TP.code.label('parent_code')
                ]).where(T1.parent_id == TP.id),
                select([
                    T1.code.label('code'),
                    expression.literal_column("null as parent_val1"),
                    expression.literal_column("null as parent_code")
                ]).where(T1.parent_id.is_(None))
            ).alias()
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2'),
                            subquery.c.parent_val1.label('parent_val1'),
                            subquery.c.parent_code.label('parent_code')])
            query = query.where(subquery.c.code == T1.code)
            query = query.where(subquery.c.code == T2.code)
            return query


@pytest.fixture(
    scope="class",
    params=[
        view_with_relationship_on_self,
        view_with_relationship_on_self_2,
    ]
)
def registry_view_with_relationship_on_self(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks(
        [], request.param)
    request.addfinalizer(registry.close)
    parent = registry.T1.insert(code='test1', val=1)
    registry.T2.insert(code='test1', val=2)
    registry.T1.insert(code='test2', val=3, parent=parent)
    registry.T2.insert(code='test2', val=4)
    return registry


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason="No sub query")
class TestViewWithRelationShipOnSelf:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_view_with_relationship_on_self):
        transaction = registry_view_with_relationship_on_self.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_ok(self, registry_view_with_relationship_on_self):
        registry = registry_view_with_relationship_on_self
        TestView = registry.TestView
        v1 = TestView.query().filter(TestView.code == 'test1').first()
        v2 = TestView.query().filter(TestView.code == 'test2').first()
        assert v1.val1 == 1
        assert v1.val2 == 2
        assert v1.parent is None
        assert v2.val1 == 3
        assert v2.val2 == 4
        assert v2.parent.code == v1.code


def simple_view_with_same_table_by_declaration_model():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            T2 = cls.anyblok.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)

    @register(Model, factory=ViewFactory, tablename=Model.TestView)
    class TestView2:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


def simple_view_with_same_table_by_name():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            T2 = cls.anyblok.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)

    @register(Model, factory=ViewFactory, tablename='testview')
    class TestView2:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


def simple_view_with_same_table_by_inherit():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            T2 = cls.anyblok.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)

    @register(Model, factory=ViewFactory)
    class TestView2(Model.TestView):
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


@pytest.fixture(
    scope="class",
    params=[
        simple_view_with_same_table_by_declaration_model,
        simple_view_with_same_table_by_name,
        simple_view_with_same_table_by_inherit,
    ]
)
def registry_view_with_inheritance(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks(
        [], request.param)
    request.addfinalizer(registry.close)
    registry.T1.insert(code='test1', val=1)
    registry.T2.insert(code='test1', val=2)
    registry.T1.insert(code='test2', val=3)
    registry.T2.insert(code='test2', val=4)
    return registry


class TestViewWithInheritance:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_view_with_inheritance):
        transaction = registry_view_with_inheritance.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_ok(self, registry_view_with_inheritance):
        registry = registry_view_with_inheritance
        TestView = registry.TestView
        TestView2 = registry.TestView2
        assert registry.TestView.__view__ == registry.TestView2.__view__
        v1 = TestView.query().filter(TestView.code == 'test1').first()
        v2 = TestView.query().filter(TestView2.code == 'test1').first()
        assert v1.val1 == v2.val1
        assert v1.val2 == v2.val2


def simple_view_without_primary_key():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String()
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.anyblok.T1
            T2 = cls.anyblok.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T1.val.label('val2')])
            return query.where(T1.code == T2.code)


def simple_view_without_view_declaration():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, factory=ViewFactory)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


@pytest.fixture(
    scope="class",
    params=[
        simple_view_without_primary_key,
        simple_view_without_view_declaration,
    ]
)
def registry_view_with_exception(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks([], request.param)
    request.addfinalizer(registry.close)
    return registry


class TestViewWithException:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_view_with_exception):
        transaction = registry_view_with_exception.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    @pytest.mark.xfail(raises=ViewException)
    def test_ok(self, registry_view_with_exception):
        # view is not create, this state must fail
        pass



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from os.path import join
from anyblok import config
from anyblok.config import (
    Configuration,
    add_configuration_file,
    add_plugins,
    add_database,
    add_install_bloks,
    add_uninstall_bloks,
    add_update_bloks,
    add_interpreter,
    add_schema,
    add_doc,
    add_logging,
    add_install_or_update_bloks,
    ConfigurationException,
    AnyBlokActionsContainer,
    ConfigOption,
    AnyBlokPlugin,
    define_preload_option,
    get_url,
    get_db_name,
    is_none,
    cast_value,
    nargs_type,
)
from anyblok.testing import tmp_configuration
from sqlalchemy.engine.url import make_url


old_getParser = config.getParser


def fnct_configuration(parser, default):
    default.update({'test': None})


def fnct_other_configuration(parser, default):
    default.update({'test': None})


def MockPluginFnct():
    pass


class MockPluginClass:
    pass


class MockArgParseArguments:

    def __init__(self, configfile=None, args=None, kwargs=None):
        if configfile:
            cfile = join('/'.join(__file__.split('/')[:-1]), configfile)
            self.configfile = cfile
            if kwargs is None:
                kwargs = {}

            kwargs['configfile'] = configfile
        else:
            self.configfile = None
        self.args = args
        self.kwargs = kwargs
        self.logging_level = None
        self.logging_configfile = None
        self.json_logging_configfile = None
        self.yaml_logging_configfile = None

    def _get_args(self):
        return self.args

    def _get_kwargs(self):
        if self.kwargs is None:
            return tuple()
        return self.kwargs.items()


class MockArguments:

    configfile = None
    logging_level = None
    logging_configfile = None
    json_logging_configfile = None
    yaml_logging_configfile = None
    vals = {'var1': 'val1',
            'var2': 'val2',
            'var3': 'val3'}

    def _get_kwargs(self):
        return self.vals.items()

    def _get_args(self):
        return False


class MockArgumentValue:

    def __init__(self, **kwargs):
        self.default = None
        self.type = str
        self.__dict__.update(kwargs)


class MockArgumentParser:

    def __init__(self, *args, **kwargs):
        self.args = MockArguments()

    def add_argument_group(self, *args, **kwargs):
        return MockArgumentParser()

    def parse_args(self, *args, **kwargs):
        return self.args

    def add_argument(self, *args, **kwargs):
        return MockArgumentValue(**kwargs)

    def set_defaults(self, *args, **kwargs):
        pass


@pytest.fixture(scope="class")
def protect_configuration(request):
    def getParser(*args, **kwargs):
        return MockArgumentParser()

    config.getParser = getParser
    old_configuration = Configuration.configuration.copy()
    old_groups = Configuration.groups.copy()
    old_labels = Configuration.labels.copy()

    def reset():
        config.getParser = old_getParser
        Configuration.configuration = old_configuration
        Configuration.groups = old_groups
        Configuration.labels = old_labels

    request.addfinalizer(reset)


class TestConfiguration:

    @pytest.fixture(autouse=True)
    def reset_conf(self, protect_configuration):
        Configuration.groups = {}
        Configuration.labels = {}
        Configuration.configuration = {}

    def assertAdded(self, group, label=None, function_=None):
        assert Configuration.groups[group] == [function_]

        if label:
            assert Configuration.labels[group] == label

    def test_add(self):
        Configuration.add('new-group', function_=fnct_configuration)
        self.assertAdded('new-group', function_=fnct_configuration)

    def test_add_with_label(self):
        Configuration.add(
            'new-group', label="One label", function_=fnct_configuration)
        self.assertAdded(
            'new-group', label="One label", function_=fnct_configuration)

    def test_add_decorator(self):

        @Configuration.add('new-group')
        def fnct(parser, default):
            default.update({'other test': True})

        self.assertAdded('new-group', function_=fnct)

    def test_has(self):
        assert Configuration.has('option') is False
        Configuration.configuration['option'] = ConfigOption('option', str)
        assert Configuration.has('option') is True

    def test_get(self):
        option = 'My option'
        Configuration.configuration['option'] = ConfigOption(option, str)
        res = Configuration.get('option')
        assert option == res

    def test_fnct_plugins_config(self):
        option = 'anyblok.tests.test_config:MockPluginFnct'
        Configuration.configuration['option'] = ConfigOption(
            option, AnyBlokPlugin)
        res = Configuration.get('option')
        assert MockPluginFnct is res

    def test_class_plugins_config(self):
        option = 'anyblok.tests.test_config:MockPluginClass'
        Configuration.configuration['option'] = ConfigOption(
            option, AnyBlokPlugin)
        res = Configuration.get('option')
        assert MockPluginClass is res

    def test_wrong_plugins_config(self):
        option = 'anyblok.tests.test_config:MockPluginWrong'
        with pytest.raises(ImportError):
            Configuration.configuration['option'] = ConfigOption(
                option, AnyBlokPlugin)

    def test_update(self):
        Configuration.update(one_option=1)
        assert Configuration.get('one_option') == 1

    def test_update2(self):
        Configuration.update(dict(one_option=1))
        assert Configuration.get('one_option') == 1

    def test_set_str(self):
        Configuration.set('value', '1')
        assert Configuration.configuration['value'].type == str
        assert Configuration.get('value') == '1'

    def test_set_int(self):
        Configuration.set('value', 1)
        assert Configuration.configuration['value'].type == int
        assert Configuration.get('value') == 1

    def test_set_float(self):
        Configuration.set('value', 1.)
        assert Configuration.configuration['value'].type == float
        assert Configuration.get('value') == 1.

    def test_set_list(self):
        Configuration.set('value', [1])
        assert Configuration.configuration['value'].type == list
        assert Configuration.get('value') == [1]

    def test_set_tuple(self):
        Configuration.set('value', (1,))
        assert Configuration.configuration['value'].type == tuple
        assert Configuration.get('value') == (1,)

    def test_set_dict(self):
        Configuration.set('value', {'a': 1})
        assert Configuration.configuration['value'].type == dict
        assert Configuration.get('value') == {'a': 1}

    def test_get_use_default_value(self):
        option = 'My option by default'
        res = Configuration.get('option', option)
        assert option == res

    def check_url(self, url, wanted_url):
        wanted_url = make_url(wanted_url)
        for x in ('drivername', 'host', 'port', 'username', 'password',
                  'database'):
            assert getattr(url, x) == getattr(wanted_url, x)

    def test_is_none_1(self):
        assert is_none(str, None) is True

    def test_is_none_2(self):
        assert is_none(str, 'None') is True

    def test_is_none_3(self):
        assert is_none(float, '') is True

    def test_is_none_4(self):
        assert is_none(str, '') is False

    def test_cast_value_1(self):
        assert cast_value(str, None) is None

    def test_cast_value_2(self):
        assert cast_value(None, 1) == 1

    def test_cast_value_3(self):
        assert cast_value(bool, 'true') is True

    def test_cast_value_4(self):
        assert cast_value(bool, 'false') is False

    def test_cast_value_5(self):
        assert cast_value(bool, 1) is True

    def test_nargs_type_1(self):
        assert nargs_type('test', 1, str)('test') == ['test']

    def test_nargs_type_2(self):
        assert nargs_type('test', 2, str)('foo\nbar') == ['foo', 'bar']

    def test_nargs_type_3(self):
        assert nargs_type('test', 2, str)('foo,bar') == ['foo', 'bar']

    def test_nargs_type_4(self):
        with pytest.raises(ConfigurationException):
            nargs_type('test', 2, str)(1)

    def test_nargs_type_5(self):
        assert nargs_type('test', 1, str)('foo\nbar') == ['foo']

    def test_nargs_type_6(self):
        assert nargs_type('test', '*', str)('foo\nbar') == ['foo', 'bar']

    def test_set_defaults(self):
        parser = self.get_parser()
        with pytest.raises(KeyError):
            parser.set_defaults(foo='bar')

    def test_AnyBlokPlugin_1(self):
        def foo():
            pass

        assert AnyBlokPlugin(foo) is foo

    def test_update_1(self):
        with pytest.raises(ConfigurationException) as e:
            Configuration.update({}, {})

        assert e.match('Too many args. Only one expected')

    def test_update_2(self):
        with pytest.raises(ConfigurationException) as e:
            Configuration.update(1)

        assert e.match('Wrong args type. Dict expected')

    def test_get_db_name_by_config_name(self):
        with tmp_configuration(db_name='anyblok',
                               db_driver_name='postgres',
                               db_host='localhost',
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            assert get_db_name() == 'anyblok'

    def test_get_db_name_by_config_url(self):
        with tmp_configuration(db_name=None,
                               db_driver_name=None,
                               db_host=None,
                               db_url='postgres://localhost/anyblok',
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            assert get_db_name() == 'anyblok'

    def test_get_db_name_ko(self):
        with tmp_configuration(db_name=None,
                               db_driver_name=None,
                               db_host=None,
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            with pytest.raises(ConfigurationException):
                get_db_name()

    def test_get_url(self):
        with tmp_configuration(db_name='anyblok',
                               db_driver_name='postgres',
                               db_host='localhost',
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            url = get_url()
            self.check_url(url, 'postgres://localhost/anyblok')

    def test_get_url2(self):
        with tmp_configuration(db_name='anyblok',
                               db_driver_name='postgres',
                               db_host='localhost',
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            url = get_url(db_name='anyblok2')
            self.check_url(url, 'postgres://localhost/anyblok2')

    def test_get_url3(self):
        with tmp_configuration(db_url='postgres:///anyblok',
                               db_name=None,
                               db_driver_name=None,
                               db_host=None,
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            url = get_url()
            self.check_url(url, 'postgres:///anyblok')

    def test_get_url4(self):
        with tmp_configuration(db_url='postgres:///anyblok',
                               db_name='anyblok2',
                               db_driver_name=None,
                               db_host=None,
                               db_user_name='jssuzanne',
                               db_password='secret',
                               db_port=None):
            url = get_url()
            self.check_url(url, 'postgres://jssuzanne:secret@/anyblok2')

    def test_get_url5(self):
        with tmp_configuration(db_url='postgres:///anyblok',
                               db_name='anyblok2',
                               db_driver_name=None,
                               db_host=None,
                               db_user_name='jssuzanne',
                               db_password='secret',
                               db_port=None):
            url = get_url(db_name='anyblok3')
            self.check_url(url, 'postgres://jssuzanne:secret@/anyblok3')

    def test_get_url_without_drivername(self):
        with tmp_configuration(db_name=None,
                               db_driver_name=None,
                               db_host=None,
                               db_user_name=None,
                               db_password=None,
                               db_port=None):
            with pytest.raises(ConfigurationException):
                get_url()

    def test_remove(self):
        Configuration.add('new-group', function_=fnct_configuration)
        Configuration.remove('new-group', function_=fnct_configuration)
        assert Configuration.groups['new-group'] == []

    def test_remove_more_function(self):
        Configuration.add('new-group', function_=fnct_configuration)
        Configuration.add('new-group', function_=fnct_other_configuration)
        Configuration.remove('new-group', function_=fnct_configuration)
        assert Configuration.groups['new-group'] == [fnct_other_configuration]

    def test_remove_label(self):
        Configuration.add(
            'new-group', label="One label", function_=fnct_configuration)
        Configuration.remove_label('new-group')
        with pytest.raises(KeyError):
            Configuration.labels['AnyBlok']['new-group']

    def test_load_without_configuration_groupes(self):
        assert Configuration.load('default') is None

    def test_empty_parse_option(self):
        args = MockArgParseArguments()
        Configuration.parse_options(args)
        assert Configuration.configuration == {}

    def assertConfig(self, kwargs):
        config = {x: y.get() for x, y in Configuration.configuration.items()}
        assert config == kwargs

    def test_parse_option(self):
        kwargs = {'test': 'value'}
        args = MockArgParseArguments(configfile="mock_configuration_file.cfg",
                                     kwargs=kwargs)
        Configuration.parse_options(args)
        kwargs.update({
            'db_name': 'anyblok',
            'db_driver_name': 'postgres',
            'db_user_name': '',
            'db_password': '',
            'db_host': 'localhost',
            'db_port': '',
        })
        self.assertConfig(kwargs)

    def test_parse_option_configuration(self):
        args = MockArgParseArguments(configfile="mock_configuration_file.cfg")
        Configuration.parse_options(args)
        self.assertConfig({
            'db_name': 'anyblok',
            'db_driver_name': 'postgres',
            'db_user_name': '',
            'db_password': '',
            'db_host': 'localhost',
            'db_port': '',
            'configfile': 'mock_configuration_file.cfg',
        })

    def test_parse_option_configuration_with_extend(self):
        args = MockArgParseArguments(
            configfile="mockblok/mock_configuration_file.cfg")
        Configuration.parse_options(args)
        self.assertConfig({
            'db_name': 'anyblok',
            'db_driver_name': 'postgres',
            'db_user_name': '',
            'db_password': '',
            'db_host': 'localhost',
            'db_port': '',
            'configfile': 'mockblok/mock_configuration_file.cfg',
        })

    def test_parse_option_kwargs(self):
        kwargs = {'test': 'value'}
        args = MockArgParseArguments(kwargs=kwargs)
        Configuration.parse_options(args)
        self.assertConfig(kwargs)

    def test_parse_option_args(self):
        args = ('test',)
        args = MockArgParseArguments(args=args)
        with pytest.raises(ConfigurationException):
            Configuration.parse_options(args)

    def test_load_with_configuration_groupes(self):
        Configuration.load('default', configuration_groups=['install-bloks'])
        self.assertConfig(MockArguments.vals)

    def test_load_with_bad_configuration_groupes(self):
        Configuration.load('default', configuration_groups=['bad-groups'])
        self.assertConfig(MockArguments.vals)

    def get_parser(self):

        class Parser(AnyBlokActionsContainer, MockArgumentParser):
            pass

        return Parser()

    def test_add_deprecated_argument(self):
        parser = self.get_parser()
        with pytest.warns(DeprecationWarning) as record:
            parser.add_argument(
                '--value', dest='value', deprecated="test deprecated")
            assert str(record.list[0].message) == 'test deprecated'

        with pytest.warns(DeprecationWarning) as record:
            Configuration.get('value')
            assert str(record.list[0].message) == 'test deprecated'

    def test_add_removed_argument_1(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', removed=True)
        with pytest.raises(ConfigurationException):
            Configuration.set('value', 1)

    def test_add_removed_argument_2(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', removed=True)
        with pytest.raises(ConfigurationException):
            Configuration.get('value')

    def test_add_removed_argument_3(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value',
                            default=1, removed=True)

    def test_add_argument_str(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', default='1')
        assert Configuration.configuration['value'].type == str
        assert Configuration.get('value') == '1'

    def test_add_argument_int(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', type=int, default=1)
        assert Configuration.configuration['value'].type == int
        assert Configuration.get('value') == 1

    def test_add_argument_float(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', type=float, default=1)
        assert Configuration.configuration['value'].type == float
        assert Configuration.get('value') == 1.

    def test_add_argument_list(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', nargs="+", default='1, 2')
        assert Configuration.get('value') == ['1', '2']

    def test_default_str(self):
        parser = self.get_parser()
        parser.add_argument('--value', dest='value', default='1')
        parser.set_defaults(value='2')
        assert Configuration.get('value') == '2'

    def test_add_application_properties(self):
        assert Configuration.applications.get(
            'test_add_application_properties') is None
        Configuration.add_application_properties(
            'test_add_application_properties', ['logging'],
            description='Just a test')
        assert Configuration.applications.get(
            'test_add_application_properties') is not None
        assert (
            Configuration.applications['test_add_application_properties'] ==
            {
                'configuration_groups': [
                    'config', 'database', 'plugins', 'logging'],
                'description': 'Just a test'
            })

    def test_add_application_properties_and_load_it(self):
        Configuration.add_application_properties(
            'test_add_application_properties', ['logging'])
        parser = MockArgumentParser()

        # add twice to check doublon
        Configuration.add('logging', function_=add_logging, label='Logging')
        Configuration.add('logging', function_=add_logging, label='Logging')

        Configuration.add('database', function_=add_database, label='Database')
        Configuration.add('config', function_=add_configuration_file)
        Configuration.add('install-bloks', function_=add_install_bloks)
        Configuration._load(parser, ['config', 'database', 'logging'])

    def test_initialize_logging(self):
        Configuration.add('logging', function_=add_logging, label='Logging')
        Configuration.set('logging_level', 'DEBUG')
        Configuration.set('logging_level_qualnames', ['test'])
        Configuration.initialize_logging()


@pytest.fixture(scope="function")
def parser():
    return MockArgumentParser()


@pytest.fixture(scope="function")
def group(parser):
    return parser.add_argument_group('label')


class TestConfigurationOption:
    function = {
        'add_configuration_file': add_configuration_file,
        'add_plugins': add_plugins,
        'add_database': add_database,
        'add_install_bloks': add_install_bloks,
        'add_uninstall_bloks': add_uninstall_bloks,
        'add_update_bloks': add_update_bloks,
        'add_interpreter': add_interpreter,
        'add_schema': add_schema,
        'add_doc': add_doc,
        'define_preload_option': define_preload_option,
        'add_logging': add_logging,
        'add_install_or_update_bloks': add_install_or_update_bloks,
    }

    def test_add_configuration_file(self, parser):
        self.function['add_configuration_file'](parser)

    def test_add_plugins(self, group):
        self.function['add_plugins'](group)

    def test_add_database(self, group):
        self.function['add_database'](group)

    def test_add_install_bloks(self, parser):
        self.function['add_install_bloks'](parser)

    def test_add_uninstall_bloks(self, parser):
        self.function['add_uninstall_bloks'](parser)

    def test_add_update_bloks(self, parser):
        self.function['add_update_bloks'](parser)

    def test_add_interpreter(self, parser):
        self.function['add_interpreter'](parser)

    def test_add_schema(self, group):
        self.function['add_schema'](group)

    def test_add_doc(self, group):
        self.function['add_doc'](group)

    def test_define_preload_option(self, parser):
        self.function['define_preload_option'](parser)

    def test_add_logging(self, parser):
        self.function['add_logging'](parser)

    def test_add_install_or_update_bloks(self, parser):
        self.function['add_install_or_update_bloks'](parser)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#    Copyright (C) 2021 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in
from anyblok.column import Integer as Int, String as Str
from anyblok.migration import MigrationException
from contextlib import contextmanager
from sqlalchemy import (
    MetaData, Table, Column, Integer, String, CheckConstraint, ForeignKey,
    text)
from anyblok import Declarations
from anyblok.config import Configuration
from .conftest import init_registry, drop_database, create_database, reset_db
from anyblok.common import naming_convention
from .test_column import simple_column
from anyblok.testing import tmp_configuration


@pytest.fixture(scope="module")
def clean_db(request, configuration_loaded):

    def clean():
        url = Configuration.get('get_url')()
        drop_database(url)
        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)

    if sgdb_in(['MySQL', 'MariaDB']):
        clean()

    request.addfinalizer(clean)


@contextmanager
def cnx(registry):
    cnx = registry.migration.conn
    try:
        yield cnx
    except Exception:
        cnx.execute(text("rollback"))
        raise


def add_in_registry():
    register = Declarations.register
    Model = Declarations.Model

    @register(Model)
    class Test:
        integer = Int(primary_key=True)
        other = Str()

    @register(Model)
    class TestUnique:
        integer = Int(primary_key=True)
        other = Str(unique=True)

    @register(Model)
    class TestIndex:
        integer = Int(primary_key=True)
        other = Str(index=True)

    if not sgdb_in(['MySQL', 'MariaDB', 'MsSQL']):
        @register(Model)
        class TestCheck:
            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheck, cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name='test'),)

    @register(Model)
    class TestFKTarget:
        integer = Int(primary_key=True)

    @register(Model)
    class TestFK:
        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer'))


@pytest.fixture()
def registry(request, clean_db, bloks_loaded):
    registry = init_registry(add_in_registry)

    def rollback():
        for table in ('reltable', 'test', 'test2', 'othername', 'testfk',
                      'testunique', 'reltab',
                      'testfktarget',  'testcheck', 'testindex'):
            try:
                registry.migration.table(table).drop()
            except MigrationException:
                pass

        registry.close()

    request.addfinalizer(rollback)
    return registry


def add_in_registry_with_schema():
    register = Declarations.register
    Model = Declarations.Model

    @register(Model)
    class Test:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Str()

    @register(Model)
    class TestUnique:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Str(unique=True)

    @register(Model)
    class TestIndex:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Str(index=True)

    if not sgdb_in(['MySQL', 'MariaDB']):
        @register(Model)
        class TestCheck:
            __db_schema__ = 'test_db_schema'

            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheck, cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name='test'),)

    @register(Model)
    class TestFKTarget:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)

    @register(Model)
    class TestFK:
        __db_schema__ = 'test_db_schema'

        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer'))


@pytest.fixture()
def registry_db_schema(request, clean_db, bloks_loaded):
    registry = init_registry(add_in_registry_with_schema)

    def rollback():
        for table in ('reltable', 'test', 'test2', 'othername', 'testfk',
                      'testunique', 'reltab',
                      'testfktarget',  'testcheck', 'testindex'):
            try:
                registry.migration.table(table, schema='test_db_schema').drop()
            except MigrationException:
                pass

        registry.close()

    request.addfinalizer(rollback)
    return registry


class TestMigration:

    def test_detect_column_added_with_protected_table(self, registry):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_table_from_config_1(
        self, registry
    ):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_models='Model.Test'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_table_from_config_2(
        self, registry
    ):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_models=['Model.Test']):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_table_from_config_3(
        self, registry
    ):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(
            ignore_migration_for_models='Model.Test,Model.System'
        ):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_table_from_config_4(
        self, registry
    ):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(
            ignore_migration_for_models='Model.Test,Model.System'
        ):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_table_from_config_5(
        self, registry
    ):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(
            ignore_migration_for_models=(
                'Model.Test,Model.System,Model.System.Blok')
        ):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_table_from_config_6(
        self, registry
    ):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(
            ignore_migration_for_models=[
                'Model.Test', 'Model.System', 'Model.System.Blok']
        ):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_column(self, registry):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': ['other']}
        report = registry.migration.detect_changed()
        assert report.log_has("Add test.other")

    def test_detect_column_removed(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has("Drop Column test.other2")

    def test_detect_nullable_with_protected_table(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), nullable=False),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has("Alter test.other")

    def test_detect_nullable_with_protected_column(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), nullable=False),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': ['other']}
        report = registry.migration.detect_changed()
        assert not report.log_has("Alter test.other")

    def test_detect_server_default_with_protected_table(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), server_default='9.99'),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has("Alter test.other")

    def test_detect_server_default_with_protected_column(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), server_default='9.99'),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': ['other']}
        report = registry.migration.detect_changed()
        assert not report.log_has("Alter test.other")

    def test_detect_drop_anyblok_index(self, registry):
        with cnx(registry) as conn:
            conn.execute(text(
                """CREATE INDEX anyblok_ix_test__other ON test (other);"""))

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has("Drop index anyblok_ix_test__other on test")

    def test_detect_type_with_protected_table(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has("Alter test.other")

    def test_detect_type_with_protected_column(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': ['other']}
        report = registry.migration.detect_changed()
        assert not report.log_has("Alter test.other")

    def test_detect_add_foreign_key_with_protected_table(self, registry):
        with cnx(registry) as conn:
            registry.TestFK.__table__.drop(bind=conn)
            registry.TestFK.__table__ = Table(
                'testfk', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
            )
            registry.TestFK.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'testfk': True}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)")

    def test_detect_add_foreign_key_with_protected_column(self, registry):
        with cnx(registry) as conn:
            registry.TestFK.__table__.drop(bind=conn)
            registry.TestFK.__table__ = Table(
                'testfk', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
            )
            registry.TestFK.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'testfk': ['other']}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)")

    def test_detect_drop_foreign_key_with_protected_table(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)
            # anyblok_fk_test__other_on_system_blok__name

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")

    def test_detect_drop_foreign_key_with_protected_column(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)
            # anyblok_fk_test__other_on_system_blok__name

        registry.migration.ignore_migration_for = {'test': ['other']}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_unique_constraint_with_protected_table(self, registry):
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
            )
            registry.TestUnique.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'testunique': True}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add unique constraint on testunique (other)")

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_unique_constraint_with_protected_column(self,
                                                                registry):
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
            )
            registry.TestUnique.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'testunique': ['other']}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add unique constraint on testunique (other)")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Drop constraint anyblok_uq_test__other on test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_add_check_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestCheck.__table__.drop(bind=conn)
            registry.TestCheck.__table__ = Table(
                'testcheck', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
            )
            registry.TestCheck.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'testcheck': True}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Add check constraint anyblok_ck_testcheck__test on testcheck")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                CheckConstraint(
                    "other != 'test'", name='check')
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.ignore_migration_for = {'test': True}
        report = registry.migration.detect_changed()
        assert not report.log_has(
            "Drop check constraint anyblok_ck__test__check on test")


@pytest.mark.column
class TestIgnoreMigrationColumns:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_insert_columns(self):
        registry = self.init_registry(simple_column, ColumnType=Str,
                                      ignore_migration=True)
        assert registry.migration.ignore_migration_for == {'test': ['col']}


class TestIgnoreSchemaMigration:

    def test_detect_column_added_with_protected_schema_from_config_1(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_schema_from_config_2(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas=['test_db_schema']):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_schema_from_config_3(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(
            ignore_migration_for_schemas='test_db_schema,other'
        ):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_added_with_protected_schema_from_config_4(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(
            ignore_migration_for_schemas=['test_db_schema', 'other']
        ):
            report = registry.migration.detect_changed()

        assert not report.log_has("Add test.other")

    def test_detect_column_removed(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Drop Column test.other2")

    def test_detect_nullable_with_protected_table(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), nullable=False),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Alter test.other")

    def test_detect_server_default_with_protected_table(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), server_default='9.99'),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Alter test.other")

    def test_detect_drop_anyblok_index(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            conn.execute(text(
                "CREATE INDEX anyblok_ix_test__other ON "
                "test_db_schema.test (other);"))

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Drop index anyblok_ix_test__other on test")

    def test_detect_type_with_protected_table(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has("Alter test.other")

    def test_detect_add_foreign_key_with_protected_table(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.TestFK.__table__.drop(bind=conn)
            registry.TestFK.__table__ = Table(
                'testfk', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
                schema='test_db_schema'
            )
            registry.TestFK.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="FIXME: can't create foreign key")
    def test_detect_drop_foreign_key_with_protected_table(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)
            # anyblok_fk_test__other_on_system_blok__name

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        message = (
            "Drop Foreign keys on test.other => %ssystem_blok.name"
        ) % ('dbo.' if sgdb_in(['MsSQL']) else '')
        assert not report.log_has(message)

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_unique_constraint_with_protected_table(
        self, registry_db_schema
    ):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                schema='test_db_schema'
            )
            registry.TestUnique.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has(
            "Add unique constraint on testunique (other)")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_anyblok_constraint(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has(
            "Drop constraint anyblok_uq_test__other on test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_add_check_constraint(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.TestCheck.__table__.drop(bind=conn)
            registry.TestCheck.__table__ = Table(
                'testcheck', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                schema='test_db_schema'
            )
            registry.TestCheck.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has(
            "Add check constraint anyblok_ck_testcheck__test on testcheck")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_anyblok_constraint(self, registry_db_schema):
        registry = registry_db_schema
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                CheckConstraint(
                    "other != 'test'", name='check'),
                schema='test_db_schema'
            )
            registry.Test.__table__.create(bind=conn)

        with tmp_configuration(ignore_migration_for_schemas='test_db_schema'):
            report = registry.migration.detect_changed()

        assert not report.log_has(
            "Drop check constraint anyblok_ck__test__check on test")



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Georges RACINET <gracinet@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Base classes for unit/integration tests with anyblok.

This module provides :class:`BlokTestCase`, which is the main one meant for
blok tests, and :class:`DBTestCase`, whose primary purpose is to test anyblok
itself, in so-called "framework tests".
"""

import unittest
import os
from anyblok.config import Configuration
from anyblok.registry import RegistryManager
from anyblok.blok import BlokManager
from anyblok.environment import EnvironmentManager
import sqlalchemy
from sqlalchemy.orm import clear_mappers
from sqlalchemy import event
from sqlalchemy_utils.functions import database_exists, create_database, orm
from copy import copy
from testfixtures import LogCapture as LC
from contextlib import contextmanager
from logging import getLogger, DEBUG, INFO, WARNING, ERROR, CRITICAL
from .common import sgdb_in as sgdb_in_, DATABASES_CACHED
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)

logger = getLogger(__name__)


class MockParser:

    def _get_kwargs(self):
        return []

    def _get_args(self):
        return False


def load_configuration():
    load_init_function_from_entry_points(unittest=True)
    Configuration.load_config_for_test()
    Configuration.parse_options(MockParser())
    configuration_post_load(unittest=True)


def drop_database(url):
    url = copy(sqlalchemy.engine.url.make_url(url))
    database = url.database
    if url.drivername.startswith('postgresql'):
        url.database = 'postgres'
    elif not url.drivername.startswith('sqlite'):
        url.database = None

    engine = sqlalchemy.create_engine(url)
    if engine.dialect.name == 'sqlite' and url.database != ':memory:':
        os.remove(url.database)
    else:
        text = 'DROP DATABASE {0}'.format(orm.quote(engine, database))
        cnx = engine.connect()
        cnx.execute("ROLLBACK")
        cnx.execute(text)
        cnx.execute("commit")
        cnx.close()


@contextmanager
def tmp_configuration(**values):
    """Add Configuration value only in the contextmanager
    ::

        with TestCase.Configuration(db_name='a db name'):
            assert Configuration.get('db_name') == 'a db name'

    :param **values: values to update
    """
    old_values = {
        x: Configuration.get(x)
        for x in values.keys()
    }
    Configuration.update(**values)
    try:
        yield
    finally:
        Configuration.update(**old_values)


class TestCase(unittest.TestCase):
    """Common helpers, not meant to be used directly."""

    _transaction_case_teared_down = False

    @classmethod
    def init_configuration_manager(cls, **env):
        """ Initialise the configuration manager with environ variable
        to launch the test

        .. warning::

            For the moment we not use the environ variable juste constante

        :param env: add another dict to merge with environ variable
        """
        db_name = Configuration.get('db_name') or 'test_anyblok'
        db_driver_name = Configuration.get('db_driver_name') or 'postgresql'
        env.update({
            'db_name': db_name,
            'db_driver_name': db_driver_name,
        })
        Configuration.update(**env)

    @classmethod
    def createdb(cls, keep_existing=False, with_demo=False):
        """Create the database specified in configuration.

        ::

            cls.init_configuration_manager()
            cls.createdb()

        :param keep_existing: If false drop the previous db before create it
        """
        url = Configuration.get('get_url')()
        db_template_name = Configuration.get('db_template_name', None)
        if database_exists(url):
            if keep_existing:
                return False

            drop_database(url)

        create_database(url, template=db_template_name)

        registry = RegistryManager.get(Configuration.get('db_name', None))
        if registry is None:
            return

        registry.System.Parameter.set(
            "with-demo", Configuration.get('with_demo', with_demo))

        return True

    @classmethod
    def dropdb(cls):
        """Drop the database specified in configuration.

        ::

            cls.init_configuration_manager()
            cls.dropdb()

        """
        url = Configuration.get('get_url')()
        if database_exists(url):
            drop_database(url)

    @classmethod
    def additional_setting(cls):
        return dict(unittest=True)

    @classmethod
    def getRegistry(cls):
        """Return the registry for the test database.

        This assumes the database is created, and the registry has already
        been initialized::

            registry = self.getRegistry()

        :rtype: registry instance
        """
        additional_setting = cls.additional_setting()
        return RegistryManager.get(Configuration.get('db_name'),
                                   **additional_setting)

    def setUp(self):
        super(TestCase, self).setUp()
        self.addCleanup(self.callCleanUp)

    def callCleanUp(self):
        if not self._transaction_case_teared_down:
            self.cleanUp()

    def cleanUp(self):
        self.tearDown()

    def tearDown(self):
        """ Roll back the session """
        super(TestCase, self).tearDown()
        self._transaction_case_teared_down = True

    Configuration = tmp_configuration


class DBTestCase(TestCase):
    """Base class for tests that need to work on an empty database.

    .. warning::

        The database is created and dropped with each unit test

    For instance, this is the one used for Field, Column, RelationShip, and
    more generally core framework tests.

    The drawback of using this base class is that tests will be slow. The
    advantage is ultimate test isolation.

    Sample usage::

        from anyblok.tests.testcase import DBTestCase


        def simple_column(ColumnType=None, **kwargs):

            @Declarations.register(Declarations.Model)
            class Test:

                id = Declarations.Column.Integer(primary_key=True)
                col = ColumnType(**kwargs)


        class TestColumns(DBTestCase):

            def test_integer(self):
                Integer = Declarations.Column.Integer

                registry = self.init_registry(simple_column,
                                              ColumnType=Integer)
                test = registry.Test.insert(col=1)
                self.assertEqual(test.col, 1)

    """

    blok_entry_points = ('bloks',)
    """setuptools entry points to load blok """

    @classmethod
    def setUpClass(cls):
        """ Intialialise the configuration manager """

        super(DBTestCase, cls).setUpClass()
        cls.init_configuration_manager()
        if cls.createdb(keep_existing=True):
            BlokManager.load(entry_points=('bloks', 'test_bloks'))
            registry = cls.getRegistry()
            registry.commit()
            registry.close()
            BlokManager.unload()

    def setUp(self):
        """ Create a database and load the blok manager """
        self.registry = None
        super(DBTestCase, self).setUp()
        BlokManager.load(entry_points=self.blok_entry_points)

    def tearDown(self):
        """ Clear the registry, unload the blok manager and  drop the database
        """
        if self.registry:
            self.registry.close()

        RegistryManager.clear()
        BlokManager.unload()
        clear_mappers()
        super(DBTestCase, self).tearDown()

    def init_registry(self, function, **kwargs):
        """ call a function to filled the blok manager with new model

        :param function: function to call
        :param kwargs: kwargs for the function
        :rtype: registry instance
        """
        return self.init_registry_with_bloks(None, function, **kwargs)

    def init_registry_with_bloks(self, bloks, function, **kwargs):
        """ call a function to filled the blok manager with new model and
        bloks to install

        :param bloks: list of blok's names
        :param function: function to call
        :param kwargs: kwargs for the function
        :rtype: registry instance
        """
        if bloks is None:
            bloks = []

        if isinstance(bloks, tuple):
            bloks = list(bloks)

        if 'anyblok-test' not in bloks:
            bloks.append('anyblok-test')

        from copy import deepcopy
        loaded_bloks = deepcopy(RegistryManager.loaded_bloks)
        if function is not None:
            EnvironmentManager.set('current_blok', 'anyblok-test')
            try:
                function(**kwargs)
            finally:
                EnvironmentManager.set('current_blok', None)

        try:
            self.registry = registry = self.__class__.getRegistry()
            if bloks:
                registry.upgrade(install=bloks)
        finally:
            RegistryManager.loaded_bloks = loaded_bloks

        return registry

    def reload_registry(self, registry, function, **kwargs):
        """ call a function to filled the blok manager with new model and
        before reload the registry

        :param bloks: list of blok's names
        :param function: function to call
        :param kwargs: kwargs for the function
        :rtype: registry instance
        """
        from copy import deepcopy
        loaded_bloks = deepcopy(RegistryManager.loaded_bloks)
        if function is not None:
            EnvironmentManager.set('current_blok', 'anyblok-test')
            try:
                function(**kwargs)
            finally:
                EnvironmentManager.set('current_blok', None)

        try:
            registry.reload()
        finally:
            RegistryManager.loaded_bloks = loaded_bloks

        return registry


class BlokTestCase(unittest.TestCase):
    """Base class for tests meant to run on a preinstalled database.

    The tests written with this class don't need to start afresh on a new
    database, and therefore run much faster than those inheriting
    :class:`DBTestCase`. Instead, they expect the tested bloks to be already
    installed and up to date.

    The session gets rollbacked after each test.

    Such tests are appropriate for a typical blok developper workflow:

    * create and install the bloks once
    * run the tests of the blok under development repeatedly
    * upgrade the bloks in database when needed (schema change, update of
      dependencies)

    They are also appropriate for on the fly testing while installing the
    bloks: the tests of each blok get run on the database state they expect,
    before dependent (downstream) bloks, that could. e.g., alter the database
    schema, get themselves installed.
    This is useful to test a whole stack at once using only one
    database (typically in CI bots).

    Sample usage::

        from anyblok.tests.testcase import BlokTestCase


        class MyBlokTest(BlokTestCase):

            def test_1(self):
                # access to the registry by ``self.registry``
                ...

    """

    _transaction_case_teared_down = False
    registry = None
    """The instance of :class:`anyblok.registry.Registry`` to use in tests.

    The ``session_commit()`` method is disabled to avoid side effects from one
    test to the other.
    """

    @classmethod
    def additional_setting(cls):
        return dict(unittest=True)

    @classmethod
    def setUpClass(cls):
        """ Initialize the registry.
        """
        super(BlokTestCase, cls).setUpClass()
        additional_setting = cls.additional_setting()
        if cls.registry is None:
            if len(BlokManager.list()) == 0:
                BlokManager.load()

            cls.registry = RegistryManager.get(Configuration.get('db_name'),
                                               **additional_setting)

        cls.registry.commit()

    def setUp(self):
        super(BlokTestCase, self).setUp()
        self.addCleanup(self.callCleanUp)
        self.registry.begin_nested()  # add SAVEPOINT

        @event.listens_for(self.registry.session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            # TODO gracinet: while working on SharedDataTestCase
            # I noticed that this could keep listening long after everything
            # has been teared down. See SharedDataTestCase for unregistration
            # example.
            if transaction.nested and not transaction._parent.nested:
                session.expire_all()
                session.begin_nested()

    def callCleanUp(self):
        if not self._transaction_case_teared_down:
            self.cleanUp()

    def cleanUp(self):
        self.tearDown()

    def tearDown(self):
        """ Roll back the session """
        super(BlokTestCase, self).tearDown()
        try:
            self.registry.System.Cache.invalidate_all()
        except sqlalchemy.exc.InvalidRequestError:
            pass
        finally:
            self.registry.rollback()
            self.registry.session.close()

        self._transaction_case_teared_down = True


class SharedDataTestCase(BlokTestCase):

    @classmethod
    def setUpClass(cls):
        super(SharedDataTestCase, cls).setUpClass()
        cls.pre_data_savepoint = cls.registry.begin_nested()
        try:
            cls.setUpSharedData()
        except Exception:
            cls.tearDownClass()
            raise

    @classmethod
    def setUpSharedData(cls):
        """To be implemented by concrete test classes."""

    @classmethod
    def make_case_savepoint(cls, session=None):
        if session is None:
            session = cls.registry
        cls.case_savepoint = session.begin_nested()

    def setUp(self):
        # we don't want to execute BlokTestCase.setUp(), only its parent's:
        super(BlokTestCase, self).setUp()
        # tearDown is not called in case of errors in setUp, but these are:
        self.addCleanup(self.callCleanUp)
        self.make_case_savepoint()

        @event.listens_for(self.registry.session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            if transaction is self.case_savepoint:
                session.expire_all()
                self.make_case_savepoint()

        self.savepoint_restarter = restart_savepoint

    @classmethod
    def tearDownClass(cls):
        cls.pre_data_savepoint.rollback()
        super(SharedDataTestCase, cls).tearDownClass()

    def tearDown(self):
        """Roll back the session """
        super(BlokTestCase, self).tearDown()
        self.case_savepoint.rollback()
        self.registry.System.Cache.invalidate_all()
        event.remove(self.registry.session, "after_transaction_end",
                     self.savepoint_restarter)
        self._transaction_case_teared_down = True


class LogCapture(LC):
    """Overwrite ``testfixtures.LogCapture`` to add some helper methods"""

    def get_messages(self, *levels):
        """Return the captured messages
        ::

            with LogCapture() as logs:
                messages = logs.get_messages(INFO, WARNING)

        :param *levels: list of logging.level
        :rtype: list of formated message
        """
        return [
            self.format(r)
            for r in self.records
            if (not levels or r.levelno in levels)
        ]

    def get_debug_messages(self):
        """Return only the logging.DEBUG messages"""
        return self.get_messages(DEBUG)

    def get_info_messages(self):
        """Return only the logging.INFO messages"""
        return self.get_messages(INFO)

    def get_warning_messages(self):
        """Return only the logging.WARNING messages"""
        return self.get_messages(WARNING)

    def get_error_messages(self):
        """Return only the logging.ERROR messages"""
        return self.get_messages(ERROR)

    def get_critical_messages(self):
        """Return only the logging.CRITICAL messages"""
        return self.get_messages(CRITICAL)


def skip_unless_bloks_installed(*bloks):
    """A decorator to skip a test if some Bloks aren't installed.

    In cases of soft dependency between Bloks (i.e., the dependent Blok
    has some of its features behaving differently if another Blok is installed
    without actually requiring it), it's useful to write tests that will
    be executed only if some Bloks are installed.

    Here's an example taken from Anyblok / Wms Base,
    where the ``wms-inventory`` Blok wants to take Reservation
    into account if it's installed yet doesn't
    want to introduce a hard requirement onto ``wms-reservation``

        @skip_unless_bloks_installed('wms-reservation')
        def test_choose_affected_with_reserved(self):
            # this test can now safely assume that wms-reservation is installed

    """
    def bloks_decorator(testmethod):
        def wrapped(self):
            Blok = self.registry.System.Blok
            for blok_name in bloks:
                blok = Blok.query().get(blok_name)
                if blok.state != 'installed':
                    raise unittest.SkipTest(
                        "Blok %r is not installed" % blok_name)
            return testmethod(self)
        # necessary not to be ignored by test runner
        wrapped.__name__ = testmethod.__name__
        return wrapped
    return bloks_decorator


def sgdb_in(databases):
    if not DATABASES_CACHED:
        load_configuration()

    url = Configuration.get('get_url')(db_name='')
    engine = sqlalchemy.create_engine(url)
    return sgdb_in_(engine, databases)
